# vortaro
# Copyright (C) 2017  Thomas Levine
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import defaultdict
from io import StringIO

class Alphabet(object):
    def __init__(self, alphabet):
        self._to_roman = alphabet
        self._from_roman = tuple((v,k) for (k,v) in alphabet)

    def __repr__(self):
        return 'Alphabet(%s)' % repr(self._to_roman)
    def to_roman(self, word):
        return self._convert(self._to_roman, word)
    def from_roman(self, word):
        return self._convert(self._from_roman, word)

    @staticmethod
    def _convert(alphabet, word):
        mapping = defaultdict(dict)
        for k, v in alphabet:
            if not 1 <= len(k) <= 2:
                raise ValueError('Character is too long: %s' % k)
            left, right = k if len(k) == 2 else (k, '')
            mapping[left][right] = v
        

        def write(raw, letter):
            upper = any(char.upper() == char and char.lower() != char for char in raw)
            output.write(letter.upper() if upper else letter.lower())

        buf = ''
        input = StringIO(word)
        output = StringIO()
        while True:
            char = input.read(1)
            if char == '':
                if buf:
                    write(buf, mapping[buf.lower()][''])
                break
            elif buf:
                if char.lower() in mapping[buf.lower()]:
                    write(buf + char, mapping[buf.lower()][char.lower()])
                    buf = ''
                elif char.lower() in mapping:
                    write(buf + char, mapping[buf.lower()][''])
                    buf = char
                else:
                    write(buf, mapping[buf.lower()][''])
                    output.write(char)
                    buf = ''
            else:
                if char.lower() not in mapping:
                    output.write(char)
                elif 1 < len(mapping[char.lower()]):
                    buf = char
                else:
                    write(char, mapping[char.lower()][''])
        return output.getvalue()

IDENTITY = Alphabet(())
ALPHABETS = dict(
    sr = Alphabet((
        ('а', 'a'),
        ('б', 'b'),
        ('ц', 'c'),
        ('д', 'd'),
        ('е', 'e'),
        ('ф', 'f'),
        ('г', 'g'),
        ('х', 'h'),
        ('и', 'i'),
        ('ј', 'j'),
        ('к', 'k'),
        ('л', 'l'),
        ('м', 'm'),
        ('н', 'n'),
        ('о', 'o'),
        ('п', 'p'),
        ('р', 'r'),
        ('с', 's'),
        ('т', 't'),
        ('у', 'u'),
        ('в', 'v'),
        ('з', 'z'),
        ('ш', 'š'),
        ('ж', 'ž'),
        ('ћ', 'ć'),
        ('ч', 'č'),
        ('ђ', 'dj'),
        ('џ', 'dž'),
        ('њ', 'nj'),
        ('љ', 'lj'),
    ))
)



# vortaro
# Copyright (C) 2017  Thomas Levine
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .transliterate import ALPHABETS, IDENTITY

UNDERLINE = '\033[4m'
NORMAL = '\033[0m'

def _sort_results(result):
    return (
        len(result['from_word']),
        result['part_of_speech'],
        result['from_word'],
        result['to_word'],
        result['from_lang'],
        result['to_lang'],
    )

class Table(object):
    def __init__(self, search, results=None):
        self.search = search
        self.results = results if results else []
    def __bool__(self):
        return bool(self.results)
    def add(self, row):
        if 'part_of_speech' not in row:
            row['part_of_speech'] = ''
        self.results.append(row)
    def __len__(self):
        return len(self.results)
    def sort(self):
        self.results.sort(key=_sort_results)
    def render(self, cols, rows):
        '''
        Supporting multibyte characters was a bit annoying. ::

            python3 -m dictcc lookup -from sv är 10
        '''
        if rows:
            results = self.results[:rows]
        else:
            results = self.results

        widths = _widths(results)

        tpl_line = '%%-0%ds\t%%-0%ds:%%-0%ds\t%%-0%ds:%%s' % tuple(widths)
        tpl_line = tpl_line.replace('\t', '   ')

        for result in results:
            formatted = tpl_line % (
                result['part_of_speech'],
                result['from_lang'],
                highlight(result['from_lang'], result['from_word'], self.search),
                result['to_lang'],
                result['to_word'],
            )
            yield formatted + '\n'

    def __repr__(self):
        return 'Table(search=%s, widths=%s, results=%s)' % \
            tuple(map(repr, (self.search, self.results)))

def highlight(lang, big_foreign, small_roman):
    alphabet = ALPHABETS.get(lang, IDENTITY)
    big_roman = alphabet.to_roman(big_foreign)

    if small_roman.lower() in big_roman.lower():
        left = big_roman.lower().index(small_roman.lower())
        right = left + len(small_roman)
        
        y = (
            alphabet.from_roman(big_roman[:left]),
            alphabet.from_roman(big_roman[left:right]),
            alphabet.from_roman(big_roman[right:]),
        )
        if ''.join(y) == big_foreign:
            a, b, c = y
            return a + UNDERLINE + b + NORMAL + c
    return big_foreign + NORMAL + NORMAL

def _widths(results):
    widths = [0, 0, 0, 0]
    for result in results:
        row = result['part_of_speech'], \
            result['from_lang'], result['from_word'], \
            result['to_lang'], result['to_word']
        for i, cell in enumerate(row[:-1]):
            widths[i] = max(widths[i], len(cell))
    return widths

def Stream(search, cols, result):
    widths = (16, 8, 28, 8)
    tpl_line = '%%-0%ds\t%%0%ds:%%-0%ds\t%%0%ds:%%s' % widths
    tpl_line = tpl_line.replace('\t', '   ')

    formatted = tpl_line % (
        result['part_of_speech'],
        result['from_lang'],
        highlight(result['from_lang'], result['from_word'], search),
        result['to_lang'],
        result['to_word'],
    )
    return formatted + '\n'



# vortaro
# Copyright (C) 2017  Thomas Levine
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import webbrowser
from sys import stdout, stderr
from textwrap import wrap
from functools import lru_cache
from shutil import get_terminal_size

COLUMNS, ROWS = get_terminal_size((80, 20))
PAIR = re.compile(rb'# ([A-Z]+)-([A-Z]+) vocabulary database	compiled by dict\.cc$')

def download(data_dir):
    directions = '''\
The download page will open in a web browser. Download the dictionary
of interest (as zipped text), unzip it, and put the text file inside this
directory: %s/''' % data_dir
    for line in wrap(directions, COLUMNS):
        stdout.write(line + '\n')
    stdout.write('\n')
    stdout.write('Press enter when you are ready.\n')
    input()
    webbrowser.open('https://www1.dict.cc/translation_file_request.php?l=e')

def read(path):
    with path.open('rb') as fp:
        firstline = fp.readline()[:-1]

    m = PAIR.match(firstline)
    if m:
        left_lang, right_lang = (g.decode('utf-8').lower() for g in m.groups())
    else:
        raise StopIteration

    with path.open() as fp:
        in_header = True
        for rawline in fp:
            if in_header:
                if rawline.startswith('#') or not rawline.strip():
                    continue
                else:
                    in_header = False
            try:
                left_word, right_word, pos = rawline[:-1].split('\t')
            except Exception as e:
                stderr.write('%s\n' % e)
            else:
                yield {
                    'search_phrase': _truncate(left_word),
                    'part_of_speech': pos,

                    'from_lang': left_lang,
                    'from_word': _truncate(left_word),
                    'to_lang': right_lang,
                    'to_word': right_word,
                }
                yield {
                    'search_phrase': _truncate(right_word),
                    'part_of_speech': pos,

                    'from_lang': right_lang,
                    'from_word': _truncate(right_word),
                    'to_lang': left_lang,
                    'to_word': left_word,
                }

_sep = re.compile(r' [\[{]')
@lru_cache(1)
def _truncate(x):
    return _sep.split(x)[0]



# vortaro
# Copyright (C) 2017  Thomas Levine
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
from os import makedirs
from sys import stderr
from hashlib import md5
from itertools import product
from collections import defaultdict

from .transliterate import ALPHABETS, IDENTITY

LOG_INTERVAL = 10000
MAX_PHRASE_LENGTH = 18

def add_history(data, search):
    with (data / 'history').open('a') as fp:
        fp.write('%s\t%s\n' % (search, datetime.datetime.now()))
def get_history(data):
    with (data / 'history').open('r') as fp:
        for line in fp:
            yield line.split('\t', 1)[0]

def _get_out_of_date(con, path):
    file_mtime = int(path.stat().st_mtime) # buffer against rounding errors
    db_mtime_str = con.get('dictionaries:%s' % path.absolute())
    return (not db_mtime_str) or \
        file_mtime > int(db_mtime_str.decode('ascii').split('.')[0])
def _set_up_to_date(con, path):
    con.set('dictionaries:%s' % path.absolute(), path.stat().st_mtime)
def _set_out_of_date(con, path):
    con.delete('dictionaries:%s' % path.absolute())

def languages(con):
    for l in con.sscan_iter(b'pairs'):
        yield tuple(l.decode('ascii').split(':'))

def search(con, from_langs, to_langs, query):
    if not from_langs or not to_langs:
        ls = tuple(languages(con))
        if not from_langs:
            from_langs = sorted(set(l[0] for l in ls))
        if not to_langs:
            to_langs = sorted(set(l[1] for l in ls))
    pairs = tuple(product(from_langs, to_langs))

    start = min(len(alphabet.from_roman(query)) \
                for alphabet in ALPHABETS.values())
    encoded_queries = {}
    for i in range(start, MAX_PHRASE_LENGTH+1):
        results = []
        for from_lang, to_lang in pairs:
            if from_lang in encoded_queries:
                encoded_query = encoded_queries[from_lang]
            else:
                alphabet = ALPHABETS.get(from_lang, IDENTITY)
                encoded_query = alphabet.from_roman(query).encode('utf-8')
                encoded_queries[from_lang] = encoded_query
            lang_key = (from_lang.encode('ascii'), to_lang.encode('ascii'))
            key = b'lengths:%s:%s:%d' % (lang_key + (i,))
            for word_fragment in con.sscan_iter(key):
                if encoded_query in word_fragment:
                    key = b'phrase:%s:%s:%s' % (lang_key + (word_fragment,))
                    for definition in con.sscan_iter(key):
                        results.append(_line_loads(definition))
        yield from sorted(results, key=lambda line: line['to_word'])

def index(con, formats, data):
    '''
    Build the dictionary language index.

    :param pathlib.Path data: Path to the data directory
    '''
    processed = set(x.decode('utf-8') for x in con.sscan_iter(b'processed'))
    for name, module in formats.items():
        directory = data / name
        if directory.is_dir():
            for file in directory.iterdir():
                if str(file.absolute()) in processed:
                    continue
                lines = tuple(module.read(file))
                phrases = defaultdict(set)
                pairs = set()
                for line in lines:
                    line['from_lang'] = line['from_lang'].lower()
                    line['to_lang'] = line['to_lang'].lower()
                    phrase = line.pop('search_phrase').lower()
                    key = (
                        line['from_lang'].encode('ascii'),
                        line['to_lang'].encode('ascii'),
                        min(MAX_PHRASE_LENGTH, len(phrase)),
                    )
                    phrases[key].add(phrase)
                    args = key[:2] + (phrase.encode('utf-8'),)
                    con.sadd(b'phrase:%s:%s:%s' % args, _line_dumps(line))
                for key, values in phrases.items():
                    con.sadd(b'lengths:%s:%s:%d' % key, *values)
                stderr.write('Processed %s\n' % file)
                con.sadd(b'processed', str(file.absolute()))
    for fullkey in con.keys(b'lengths:*'):
        _, from_lang, to_lang, _ = fullkey.split(b':')
        con.sadd(b'pairs', b'%s:%s' % (from_lang, to_lang))

def _line_dumps(x):
    return '\t'.join((
        x.get('part_of_speech', ''),
        x['from_word'], x['from_lang'],
        x['to_word'], x['to_lang'],
    )).encode('utf-8')
def _line_loads(x):
    y = {}
    y['part_of_speech'], y['from_word'], y['from_lang'], \
        y['to_word'], y['to_lang'] = x.decode('utf-8').split('\t')
    return y



# vortaro
# Copyright (C) 2017  Thomas Levine
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from functools import lru_cache
from pathlib import Path
from gzip import decompress
from sys import stderr, exit

from .http import simple_download

LICENSE = (
    'Please observe the copyrights CC-CEDICT.',
    '''\
CC-CEDICT is licensed under a Creative Commons Attribution-Share Alike 3.0
License (http://creativecommons.org/licenses/by-sa/3.0/). It more or less means
that you are allowed to use this data for both non-commercial and commercial
purposes provided that you: mention where you got the data from (attribution)
and that in case you improve / add to the data you will share these changes
under the same license (share alike).''',
    'https://www.mdbg.net/chinese/dictionary?page=cc-cedict',
)
URL = 'https://www.mdbg.net/chinese/export/cedict/cedict_1_0_ts_utf-8_mdbg.txt.gz'
FILENAME = Path(URL).with_suffix('.txt').name

def download(directory):
    file = (directory / FILENAME)
    if file.exists():
        stderr.write('CC-CEDICT is already downloaded.\n')
        exit(1)
    else:
        body = decompress(simple_download(URL, LICENSE, directory))
        with file.open('wb') as fp:
            fp.write(body)

def read(path):
    '''
    Read a dictionary file

    :param pathlib.Path: Dictionary file
    '''
    with path.open() as fp:
        in_header = True
        for rawline in fp:
            if in_header:
                if rawline.startswith('#'):
                    continue
                else:
                    in_header = False

            traditional, simplified, _rest = rawline[:-1].split(' ', 2)
            _pinyin, *englishes, _ = _rest.split('/')
            pinyin = _pinyin[1:-2]
            ppinyin = _render_pinyin(pinyin)

            for english in englishes:
                base = {
                    'from_lang': 'en',
                    'from_word': english,
                    'search_phrase': english,
                }
                yield dict(**base, **{
                    'to_lang': 'zh',
                    'to_word': '%s [%s]' % (traditional, ppinyin),
                })
                yield dict(**base, **{
                    'to_lang': 'zh_CN',
                    'to_word': '%s [%s]' % (simplified, ppinyin),
                })

                base = {
                    'to_lang': 'en',
                    'to_word': english,
                }
                yield dict(**base, **{
                    'search_phrase': pinyin,
                    'from_lang': 'zh',
                    'from_word': '%s [%s]' % (traditional, ppinyin),
                })
                yield dict(**base, **{
                    'search_phrase': pinyin,
                    'from_lang': 'zh_CN',
                    'from_word': '%s [%s]' % (simplified, ppinyin),
                })

TONES = {
    'a': 'āáǎàa',
    'e': 'ēéěèe',
    'i': 'īíǐìi',
    'o': 'ōóǒòo',
    'u': 'ūúǔùu',
} 

def _render_pinyin(xs):
    return ' '.join(map(_render_syllable, xs.split()))
    
@lru_cache(None)
def _render_syllable(x):
    if x[-1] not in set('12345'):
        return x

    syllable = x[:-1]
    tone = int(x[-1]) - 1
    if 'ao' in syllable:
        original = 'ao'
        replacement = TONES['a'][tone] + 'o'
    else:
        for v in 'aeo':
            if v in syllable:
                original = v
                replacement = TONES[v][tone]
                break
        else:
            for vs in ('iu', 'ui'):
                if vs in syllable:
                    original = vs
                    l, r = vs
                    replacement = l + TONES[r][tone]
                    break
            else:
                for v in 'aoeui':
                    if v in syllable:
                        original = v
                        replacement = TONES[v][tone]
                        break
                else:
                    return x
    return syllable.replace(original, replacement)



# vortaro
# Copyright (C) 2017  Thomas Levine
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from sys import stderr, exit
from functools import lru_cache

from .http import simple_download

URL = 'http://www.denisowski.org/Esperanto/ESPDIC/espdic.txt'
FILENAME = URL.split('/')[-1]
LICENSE = (
    'The purpose of the ESPDIC project is to create an electronic Esperanto-English dictionary in the form of a single Unicode (UTF8) text file with the following format :',
	'Esperanto : English',
	'(i.e. Esperanto entry, space, colon, space, English definition - one entry per line). In some cases, a semicolon is used to show different definitions for Esperanto homonyms, e.g. "kajo : conjunction; quay, wharf". This simple formatting is designed to facilitate the integration of ESPDIC into other applications (which is highly encouraged).',
	'In order to encourage its use in the Esperanto community, ESPDIC (Esperanto English Dictionary) by Paul Denisowski is licensed under a Creative Commons Attribution 3.0 Unported License. What this means is that anyone can use, transmit, or modify ESPDIC for any purpose, including commercial purposes, as long as the source is properly attributed.',
)

def download(directory):
    file = (directory / FILENAME)
    if file.exists():
        stderr.write('ESPDIC is already downloaded.\n')
        exit(1)
    else:
        body = simple_download(URL, LICENSE, directory)
        with file.open('wb') as fp:
            fp.write(body)

def read(path):
    '''
    Read a dictionary file

    :param pathlib.Path path: Dictionary file
    '''
    with path.open() as fp:
        next(fp)
        for rawline in fp:
            l, r = rawline[:-1].split(' : ')
            yield {
                'search_phrase': l,
                'part_of_speech': _part_of_speech(l[-2:]),

                'from_lang': 'eo',
                'from_word': l,
                'to_lang': 'en',
                'to_word': r,
            }
            yield {
                'search_phrase': r,
                'part_of_speech': _part_of_speech(l[-2:]),

                'from_lang': 'en',
                'from_word': r,
                'to_lang': 'eo',
                'to_word': l,
            }

POS = (
    ('noun', ('o', 'oj')),
    ('adj',  ('a', 'aj')),
    ('adv',  ('e',)),
    ('verb', ('i', 'is', 'as', 'os', 'u')),
)

@lru_cache(None)
def _part_of_speech(word):
    for pos, suffixes in POS:
        for suffix in suffixes:
            if word.endswith(suffix):
                return pos
    return ''



from textwrap import wrap
from sys import stdout, stderr, exit
from shutil import get_terminal_size

from requests import get as _get

COLUMNS, ROWS = get_terminal_size((80, 20))

def get(url):
    return _get(url,
                headers={'user-agent': 'https://pypi.python.org/pypi/vortaro'})

def simple_download(url, license, directory):
    for paragraph in license:
        for line in wrap(paragraph, COLUMNS):
            stdout.write(line + '\n')
        stdout.write('\n')
    r = get(url)
    if r.ok:
        return r.content
    else:
        stderr.write('Problem downloading %s\n' % url)
        exit(1)



# vortaro
# Copyright (C) 2017  Thomas Levine
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from shlex import split
from os import environ, makedirs
from sys import stdin, stdout, stderr, exit
from pathlib import Path
from itertools import product
from functools import partial
from collections import OrderedDict
from shutil import get_terminal_size

from redis import StrictRedis

from .render import Table, Stream
from . import (
    db,
    dictcc, cedict, espdic
)
FORMATS = OrderedDict((
    ('dict.cc', dictcc),
    ('cc-cedict', cedict),
    ('espdic', espdic),
))
DB = 7 # Redis DB number

COLUMNS, ROWS = get_terminal_size((80, 20))
DATA = Path(environ.get('HOME', '.')) / '.vortaro'

def Word(x):
    illegal = set('\t\n\r')
    if set(x).issubset(illegal):
        raise ValueError('Word contains forbidden characters.')
    else:
        return x

def download(source: tuple(FORMATS), force=False, data_dir: Path=DATA,
             redis_host='localhost', redis_port: int=6379, redis_db: int=DB):
    '''
    Download a dictionary.

    :param source: Dictionary source to download from
    :param pathlib.path data_dir: Vortaro data directory
    :param bool force: Force updating of the index
    :param redis_host: Redis host
    :param redis_port: Redis port
    :param redis_db: Redis database number
    '''
    subdir = data_dir / source
    makedirs(subdir, exist_ok=True)
    FORMATS[source].download(subdir)

    con = StrictRedis(host=redis_host, port=redis_port, db=redis_db)
    db.index(con, FORMATS, data_dir)

def index(drop=False, data_dir: Path=DATA,
          redis_host='localhost', redis_port: int=6379, redis_db: int=DB):
    '''
    Download a dictionary.

    :param pathlib.path data_dir: Vortaro data directory
    :param bool drop: Delete the existing index before indexing.
    :param redis_host: Redis host
    :param redis_port: Redis port
    :param redis_db: Redis database number
    '''
    con = StrictRedis(host=redis_host, port=redis_port, db=redis_db)
    if drop:
        con.flushdb()
    db.index(con, FORMATS, data_dir)

def languages(redis_host='localhost', redis_port: int=6379, redis_db: int=DB):
    '''
    List from-languages that have been indexed.

    :param redis_host: Redis host
    :param redis_port: Redis port
    :param redis_db: Redis database number
    '''
    con = StrictRedis(host=redis_host, port=redis_port, db=redis_db)
    for language in db.languages(con):
        print(language)

def stream(search: Word, limit: int=ROWS-2, *, width: int=COLUMNS,
           data_dir: Path=DATA,
           redis_host='localhost', redis_port: int=6379, redis_db: int=DB,
           from_langs: [str]=(), to_langs: [str]=()):
    '''
    Search for a word in the dictionaries.

    :param search: The word/fragment you are searching for
    :param limit: Maximum number of words to return
    :param width: Number of column in a line, or 0 to disable truncation
    :param from_langs: Languages the word is in, defaults to all
    :param to_langs: Languages to look for translations, defaults to all
    :param pathlib.Path data_dir: Vortaro data directory
    :param redis_host: Redis host
    :param redis_port: Redis port
    :param redis_db: Redis database number
    '''
    con = StrictRedis(host=redis_host, port=redis_port, db=redis_db)
    fmt = partial(Stream, search, width)

    i = 0
    for definition in db.search(con, from_langs, to_langs, search):
        i += 1
        stdout.write(fmt(definition))
        if i >= limit:
            break
    if i:
        db.add_history(data_dir, search)

def table(search: Word, limit: int=ROWS-2, *, width: int=COLUMNS,
          data_dir: Path=DATA,
          redis_host='localhost', redis_port: int=6379, redis_db: int=DB,
          from_langs: [str]=(), to_langs: [str]=()):
    '''
    Search for a word in the dictionaries, and format the result as a table.

    :param search: The word/fragment you are searching for
    :param limit: Maximum number of words to return
    :param width: Number of column in a line, or 0 to disable truncation
    :param from_langs: Languages the word is in, defaults to all
    :param to_langs: Languages to look for translations, defaults to all
    :param pathlib.Path data_dir: Vortaro data directory
    :param redis_host: Redis host
    :param redis_port: Redis port
    :param redis_db: Redis database number
    '''
    con = StrictRedis(host=redis_host, port=redis_port, db=redis_db)

    t = Table(search)
    for definition in db.search(con, from_langs, to_langs, search):
        t.add(definition)
        if len(t) >= limit:
            break
    t.sort()

    if t:
        db.add_history(data_dir, search)
    for row in t.render(width, limit):
        stdout.write(row)



# vortaro
# Copyright (C) 2017  Thomas Levine
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from distutils.core import setup

setup(name='vortaro',
      author='Thomas Levine',
      author_email='_@thomaslevine.com',
      description='Query bilingual dictionaries.',
      url='https://thomaslevine.com/scm/vortaro/',
      packages=['vortaro'],
      install_requires=['horetu>=0.4.1'],
      classifiers=[
          'Programming Language :: Python :: 3.6',
      ],
      version='0.3',
      license='GPL',
      entry_points = {
          'console_scripts': ['vortaro = dadaportal:dadaportal']
      },
      )
